/* ----------------------------------------------------------------------------------------------- */
/*  Copyright (c) 2014 - 2019 by Axel Kenzo, axelkenzo@mail.ru                                     */
/*                                                                                                 */
/*  Файл ak_mac.h                                                                                  */
/*  - содержит описания функций, реализующих алгоритмы итерационного сжатия                        */
/* ----------------------------------------------------------------------------------------------- */
#ifndef __AK_MAC_H__
#define __AK_MAC_H__

/* ----------------------------------------------------------------------------------------------- */
 #include <ak_oid.h>

/* ----------------------------------------------------------------------------------------------- */
/*! \brief Функция очистки контекста хеширования. */
 typedef int ( ak_function_context_clean )( ak_pointer );
/*! \brief Однораундовая функция сжатия, применяемая к одному или нескольким входным блокам. */
 typedef int ( ak_function_context_update )( ak_pointer, const ak_pointer , const size_t );
/*! \brief Функция завершения вычислений и получения конечного результата. */
 typedef int ( ak_function_context_finalize )( ak_pointer,
                                     const ak_pointer , const size_t , ak_pointer , const size_t );

/* ----------------------------------------------------------------------------------------------- */
/*! \brief Максимальный размер блока входных данных. */
 #define ak_mac_context_max_buffer_size (64)

/* ----------------------------------------------------------------------------------------------- */
/*! \brief Контекст алгоритима итерационного сжатия. */
/*! Класс предоставляет интерфейс для реализации процедруры последовательного итерационного сжатия.
    Данная процедура применяется в алгоритмах хеширования, алгоритмах выработки имитовставки и т.п.*/
/* ----------------------------------------------------------------------------------------------- */
 typedef struct mac {
  /*! \brief Размер входного блока данных (в октетах) */
   size_t bsize;
  /*! \brief Текущее количество данных во внутреннем буффере. */
   size_t length;
  /*! \brief Внутренний буффер для хранения входных данных. */
   ak_uint8 data[ ak_mac_context_max_buffer_size ];
  /*! \brief Указатель на контекст, содержащий внутреннее состояние алгоритма сжатия. */
   ak_pointer ctx;
  /*! \brief Функция очистки контекста ctx */
   ak_function_context_clean *clean;
  /*! \brief Функция обновления состояния контекста ctx  */
   ak_function_context_update *update;
  /*! \brief Функция завершения вычислений и получения конечного результата */
   ak_function_context_finalize *finalize;
 } *ak_mac;

/* ----------------------------------------------------------------------------------------------- */
/*! \brief Функция инициализации контекста начальными значениями. */
 int ak_mac_context_create( ak_mac , const size_t , ak_pointer ,
     ak_function_context_clean * , ak_function_context_update * , ak_function_context_finalize * );
/*! \brief Функция удаления контекста. */
 int ak_mac_context_destroy( ak_mac );
/*! \brief Очистка контекста сжимающего отображения. */
 int ak_mac_context_clean( ak_mac );
/*! \brief Обновление состояния контекста сжимающего отображения. */
 int ak_mac_context_update( ak_mac , const ak_pointer , const size_t );
/*! \brief Обновление состояния и вычисление результата применения сжимающего отображения. */
 int ak_mac_context_finalize( ak_mac , const ak_pointer , const size_t , ak_pointer , const size_t );
/*! \brief Применение сжимающего отображения к заданной области памяти. */
 int ak_mac_context_ptr( ak_mac , ak_pointer , const size_t , ak_pointer , const size_t );
/*! \brief Применение сжимающего отображения к заданному файлу. */
 int ak_mac_context_file( ak_mac , const char* , ak_pointer , const size_t );

#endif
/* ----------------------------------------------------------------------------------------------- */
/*                                                                                       ak_mac.h  */
/* ----------------------------------------------------------------------------------------------- */
