/* ----------------------------------------------------------------------------------------------- */
/*  Copyright (c) 2014 - 2020 by Axel Kenzo, axelkenzo@mail.ru                                     */
/*                                                                                                 */
/*  Файл ak_oid.h                                                                                  */
/*  - содержит описания функций, реализующих операции с идентификаторами криптографических         */
/*    алгоритмов и параметров                                                                      */
/* ----------------------------------------------------------------------------------------------- */
#ifndef    __AK_OID_H__
#define    __AK_OID_H__

/* ----------------------------------------------------------------------------------------------- */
 #include <libakrypt.h>

/* предварительное описание */
 typedef struct oid *ak_oid;

/* ----------------------------------------------------------------------------------------------- */
/*! \brief Структура, контролирующая функционирование криптографических объектов библиотеки. */
/*! \details Структура представляет из себя описатель класса, позволяющий создавать и уничтожать
 объекты данного класса, а также вызывать базовые функции чтения и сохранения объектов.            */
/* ----------------------------------------------------------------------------------------------- */
 typedef struct object {
 /*! \brief Размер области памяти для объекта */
  size_t size;
 /*! \brief Конструктор объекта. */
  ak_function_create_object *create;
 /*! \brief Деструктор объекта. */
  ak_function_destroy_object *destroy;
} *ak_object;

/* ----------------------------------------------------------------------------------------------- */
#define ak_object_undefined { 0, NULL, NULL }

/* ----------------------------------------------------------------------------------------------- */
/*! \brief Класс для хранения идентификаторов объектов (криптографических механизмов) и их данных. */
/*! OID (Object IDentifier) это уникальная последовательность чисел, разделенных точками.
    OID'ы могут быть присвоены любому криптографическому механизму (алгоритму,
    схеме, протоколу), а также произвольным параметрам этих механизмов.
    Использование OID'в позволяет однозначно определять тип криптографического механизма или
    значения его параметров на этапе выполнения программы, а также
    однозначно связывать данные (как правило ключевые) с алгоритмами, в которых эти данные
    используются.                                                                                  */
/* ----------------------------------------------------------------------------------------------- */
 typedef struct oid {
  /*! \brief Открытая часть структуры, содержащая доступную пользователю информацию об идентификаторе */
   struct oid_info info;
  /*! \brief Указатель на данные. */
   ak_pointer data;
  /*! \brief Структура, контролирующая поведение объекта */
   struct object func;
} *ak_oid;

/* ----------------------------------------------------------------------------------------------- */
/*! \brief Создание объекта в оперативной памяти (куче) */
 ak_pointer ak_oid_context_new_object( ak_oid );
/*! \brief Удаление объекта из кучи */
 ak_pointer ak_oid_context_delete_object( ak_oid , ak_pointer );

/* ----------------------------------------------------------------------------------------------- */
/*! \brief Поиск OID его имени. */
 ak_oid ak_oid_context_find_by_name( const char * );
/*! \brief Поиск OID по его идентификатору (строке цифр, разделенных точками). */
 ak_oid ak_oid_context_find_by_id( const char * );
/*! \brief Поиск OID по его имени или идентификатору. */
 ak_oid ak_oid_context_find_by_ni( const char * );
/*! \brief Поиск OID по указателю на даные */
 ak_oid ak_oid_context_find_by_data( ak_const_pointer  );
/*! \brief Поиск OID по типу криптографического механизма. */
 ak_oid ak_oid_context_find_by_engine( const oid_engines_t );
/*! \brief Продолжение поиска OID по типу криптографического механизма. */
 ak_oid ak_oid_context_findnext_by_engine( const ak_oid, const oid_engines_t );
/*! \brief Проверка соответствия заданного адреса корректному oid. */
 bool_t ak_oid_context_check( const ak_oid );

#endif
/* ----------------------------------------------------------------------------------------------- */
/*                                                                                       ak_oid.h  */
/* ----------------------------------------------------------------------------------------------- */
